/*
 * Copyright (c) 2014, Fabien LOISON <http://flozz.fr/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice, this
 *     list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *   * Neither the name of Fabien LOISON nor the names of its contributors may be used
 *     to endorse or promote products derived from this software without specific
 *     prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @module jsMined
 */


/**
 * Load assets.
 *
 * Cherry-picked from wonder.js <http://bzr.flogisoft.com/wonderjs/>
 *
 * @class Loader
 * @constructor
 */
var Loader = Class.$extend({

    /**
     * Image collection.
     *
     * @property _images
     * @private
     * @type Object
     */
    _images: {},

    /**
     * Sound collection.
     *
     * @property _sounds
     * @private
     * @type Object
     */
    _sounds: {},

    // Constructor
    __init__: function() {
        // New instances
        this._images = {};
        this._sounds = {};
    },

    /**
     * Load an image.
     *
     * @method loadImage
     * @param {String} imageId An unique identifier for the image.
     * @param {String} url The image url.
     */
    loadImage: function(imageId, url) {
        this._images[imageId] = {'loaded': false, 'res': new Image()};

        this._images[imageId]['res'].src = url;
        this._images[imageId]['res'].onload = function() {
            this._images[imageId]['loaded'] = true;
        }.bind(this);
    },

    /**
     * Get a loaded image.
     *
     * @method getImage
     * @param {String} imageId The unique identifier of the image.
     * @return {Image} The image.
     */
    getImage: function(imageId) {
        return this._images[imageId]['res'];
    },

    /**
     * Load a sound.
     *
     * @method loadSound
     * @param {String} soundId An unique identifier for the sound.
     * @param {String} url The image url.
     */
    loadSound: function(soundId, url) {
        this._sounds[soundId] = {
            'res': document.createElement('audio')
        };

        this._sounds[soundId]['res'].src = url;
        this._sounds[soundId]['res'].load();
    },

    /**
     * Get a loaded sound.
     *
     * @method getSound
     * @param {String} soundId The unique identifier of the sound.
     * @return {Object} The sound.
     */
    getSound: function(soundId) {
        return this._sounds[soundId]['res'];
    },

    /**
     * Wait until everything is loaded, and call the callback function.
     * @method waitLoad
     * @param {Function} callback The callback function.
     */
    waitLoad: function(callback) {
        // Check images
        for (imageId in this._images) {
            if (!this._images[imageId]['loaded']) {
                setTimeout(this.waitLoad.bind(this, callback), 20);
                return;
            }
        }

        // Check sounds
        for (soundId in this._sounds) {
            if (!this._sounds[soundId]['res'].readyState) {
                setTimeout(this.waitLoad.bind(this, callback), 20);
                return;
            }
        }

        callback();
    }
});


/**
 * Manage a game and its grid.
 *
 * @class Game
 * @constructor
 * @param canvas The canvas where the game board will be drawn.
 * @param {Number} gridWidth The grid width (optional, default=10).
 * @param {Number} gridHeight The grid height (optional, default=10).
 * @param {Number} nbMines The number of mines (optional, default=20%).
 */
var Game = Class.$extend({

    /**
     * The cells size (const).
     *
     * @property CELL_SIZE
     * @readOnly
     * @type Number
     */
    CELL_SIZE: 40,

    /**
     * @property gridWidth
     * @readOnly
     * @type Number
     */
    gridWidth: 10,

    /**
     * @property gridHeight
     * @readOnly
     * @type Number
     */
    gridHeight: 10,

    /**
     * Number of mine.
     *
     * @property nbMines
     * @readOnly
     * @type Number
     */
    nbMines: 0,

    /**
     * Number of flags on the gameboard.
     *
     * @property flags
     * @readOnly
     * @type Number
     */
    flags: 0,

    /**
     * Indicate if the party is finished (true) or not (false).
     *
     * @property gameOver
     * @readOnly
     * @type Boolean
     */
    gameOver: false,

    /**
     * Indicate if the player won (true) or not (false).
     *
     * @property gameOver
     * @readOnly
     * @type Boolean
     */
    win: false,

    /**
     * The canvas.
     *
     * @property _canvas
     * @private
     * @type HTMLElement / photonui.Canvas
     */
    _canvas: null,

    /**
     * Contain all the informations about the game board (position of mines,...).
     *
     *   * -1: mine
     *   * 0..8: water (the number represent the number of mines around the cell).
     *
     * @property _grid
     * @private
     * @type Array
     */
    _grid: [],

    /**
     * Contain information about discovered/flagged cells.
     *
     *   * -1: not discovered
     *   * 0: discovered
     *   * 1: flagged
     *
     * @property _discovered
     * @private
     * @type Array
     */
    _discovered: [],

    /**
     * Indicate if the game was initialized (_grid generated).
     *
     * @property _initialized
     * @private
     * @type Boolean
     */
    _initialized: false,


    // Constructor
    __init__: function(canvas, gridWidth, gridHeight, nbMines) {
        // Params
        this._canvas = canvas;
        if (gridWidth) this.gridWidth = gridWidth;
        if (gridHeight) this.gridHeight = gridHeight;
        if (nbMines) {
            this.nbMines = nbMines;
        }
        else {
            this.nbMines = Math.round((this.gridWidth * this.gridHeight) * .20);
        }
        // Canvas size
        this._canvas.width = this.gridWidth * this.CELL_SIZE;
        this._canvas.height = this.gridHeight * this.CELL_SIZE;
        // Init grid
        this._grid = [],
        this._discovered = [];
        for (var y=0 ; y<this.gridHeight ; y++) {
            this._grid[y] = [];
            this._discovered[y] = [];
            for (var x=0 ; x<this.gridWidth ; x++) {
                this._grid[y][x] = 0;
                this._discovered[y][x] = -1;
            }
        }
        // Draw
        this.refresh();
    },

    /**
     * Discover a cell
     *
     * @method discover
     * @param {Number} x
     * @param {Number} y
     */
    discover: function(x, y) {
        // Initialize
        if (!this._initialized) {
            this._initGame(x, y);
        }
        // Game Over
        if (this.gameOver) {
            return;
        }
        // Already discovered
        if (this._discovered[y][x] != -1) {
            return;
        }
        // Discover cell
        switch (this._grid[y][x]) {
        case -1:
            this.gameOver = true;
            this.win = false;
            break;
        case 0:
            this._discovered[y][x] = 0;
            this._expand();
            break;
        default:
            this._discovered[y][x] = 0;
            break;
        }
    },

    /**
     * Flag / Unflag a cell.
     *
     * @method toggleFlag
     * @param {Number} x
     * @param {Number} y
     */
    toggleFlag: function(x, y) {
        // Initialize
        if (!this._initialized) {
            this._initGame(x, y);
        }
        // Game Over
        if (this.gameOver) {
            return;
        }
        // Toggle the flag
        if (this._discovered[y][x] == -1) {
            this._discovered[y][x] = 1;
            this.flags++;
        }
        else if (this._discovered[y][x] == 1) {
            this._discovered[y][x] = -1;
            this.flags--;
        }
        this._checkWin();
    },

    /**
     * Refresh the canvas.
     *
     * @method refresh
     */
    refresh: function() {
        var ctx = this._canvas.getContext("2d");

        ctx.save();

        // -- Background --
        ctx.drawImage(loader.getImage("bg-water"), 0, 0);
        ctx.globalAlpha = .1;
        ctx.drawImage(loader.getImage("bg-discovered"), 0, 0);
        ctx.globalAlpha = 1;

        // -- Discovered and Flags mask --
        var discoveredMask = document.createElement("canvas");
        discoveredMask.width = this._canvas.width;
        discoveredMask.height = this._canvas.height;
        var dmctx = discoveredMask.getContext("2d");

        var flagsMask = document.createElement("canvas");
        flagsMask.width = this._canvas.width;
        flagsMask.height = this._canvas.height;
        var fmctx = flagsMask.getContext("2d");

        for (var y=0 ; y<this.gridHeight ; y++) {
            for (var x=0 ; x<this.gridWidth ; x++) {
                // Discovered
                if (this._discovered[y][x] == 0) {
                    dmctx.fillRect(
                        x * this.CELL_SIZE,
                        y * this.CELL_SIZE,
                        this.CELL_SIZE,
                        this.CELL_SIZE
                    );
                }
                // Flags
                else if (this._discovered[y][x] == 1) {
                    fmctx.fillRect(
                        x * this.CELL_SIZE,
                        y * this.CELL_SIZE,
                        this.CELL_SIZE,
                        this.CELL_SIZE
                    );
                }
            }
        }

        // -- Discovered --
        var discoveredCalc = document.createElement("canvas");
        discoveredCalc.width = this._canvas.width;
        discoveredCalc.height = this._canvas.height;
        var dcctx = discoveredCalc.getContext("2d");

        dcctx.drawImage(loader.getImage("bg-water"), 0, 0);
        dcctx.globalCompositeOperation = "destination-out";
        dcctx.drawImage(discoveredMask, 0, 0);

        // Shadow
        dmctx.globalCompositeOperation = "source-out";
        dmctx.fillRect(0, 0, discoveredMask.width, discoveredMask.height);
        stackBlurCanvasRGBA(discoveredMask, 0, 0, discoveredMask.width, discoveredMask.height, 10);
        ctx.globalAlpha = .4;
        ctx.drawImage(discoveredMask, 0, 0);
        ctx.globalAlpha = 1;

        ctx.drawImage(discoveredCalc, 0, 0);

        // -- Flags --
        var flagsCalc = document.createElement("canvas");
        flagsCalc.width = this._canvas.width;
        flagsCalc.height = this._canvas.height;
        var fcctx = flagsCalc.getContext("2d");

        fcctx.drawImage(loader.getImage("bg-flags"), 0, 0);
        fcctx.globalCompositeOperation = "destination-in";
        fcctx.drawImage(flagsMask, 0, 0);

        ctx.drawImage(flagsCalc, 0, 0);

        // Symbols
        // TODO


        // FIXME TMP
        for (var y=0 ; y<this.gridHeight ; y++) {
            for (var x=0 ; x<this.gridWidth ; x++) {
                if (this._discovered[y][x] == 0) {
                    if (this._grid[y][x] > 0) {
                        ctx.fillStyle = "#fff";
                        ctx.textAlign = "center";
                        ctx.font = "18pt Arial";
                        ctx.fillText(
                            ""+this._grid[y][x],
                            x * this.CELL_SIZE + this.CELL_SIZE/2,
                            y * this.CELL_SIZE + this.CELL_SIZE/2 + 8
                        );
                    }
                }
            }
        }

        // FIXME TMP Display mines (game over)
        ctx.fillStyle = "#111";
        if (this.gameOver && !this.win) {
            for (var y=0 ; y<this.gridHeight ; y++) {
                for (var x=0 ; x<this.gridWidth ; x++) {
                    if (this._grid[y][x] == -1) {
                        ctx.beginPath();
                        ctx.arc(
                            x*this.CELL_SIZE + this.CELL_SIZE/2,
                            y*this.CELL_SIZE + this.CELL_SIZE/2,
                            this.CELL_SIZE/2 - 5,
                            0,
                            2*Math.PI
                        );
                        ctx.fill();
                    }
                }
            }
        }

        ctx.restore();

        // FIXME DEBUG
        //ctx.save();
        //for (var y=0 ; y<this.gridHeight ; y++) {
            //for (var x=0 ; x<this.gridWidth ; x++) {
                //ctx.fillStyle = "rgba(255, 255, 255, .4)";
                //ctx.fillText(
                    //""+this._grid[y][x],
                    //x * this.CELL_SIZE + 2,
                    //y * this.CELL_SIZE + 10
                //);
            //}
        //}
        //ctx.restore();
    },

    /**
     * Initialize the game (generates mines,...)
     *
     * @method _initGame
     * @private
     * @param {Number} ix The x position of the first clicked cell.
     * @param {Number} iy The y position of the first clicked cell.
     */
    _initGame: function(ix, iy) {
        // Place mines
        var mines = this.nbMines;
        while (mines > 0) {
            var x = Math.floor(Math.random() * this.gridWidth);
            var y = Math.floor(Math.random() * this.gridHeight);
            if (this._grid[y][x] == 0 && !(x >= ix-1 && x <= ix+1 && y >= iy-1 && y <= iy+1)) {
                this._grid[y][x] = -1;
                mines--;
            }
        }
        // Count mines
        for (var y=0 ; y<this.gridHeight ; y++) {
            for (var x=0 ; x<this.gridWidth ; x++) {
                if (this._grid[y][x] == -1) continue;
                var mines = 0;
                for (var y2=Math.max(0, y-1) ; y2<=Math.min(this.gridHeight-1, y+1) ; y2++) {
                    for (var x2=Math.max(0, x-1) ; x2<=Math.min(this.gridWidth-1, x+1) ; x2++) {
                        if (this._grid[y2][x2] == -1) mines++;
                    }
                }
                this._grid[y][x] = mines;
            }
        }

        this._initialized = true;
    },

    /**
     * Expland the water zones when the player discover a cell with no mines around.
     *
     * @method _expland
     * @private
     */
    _expand: function() {
        var change = false;
        for (var y=0 ; y<this.gridHeight ; y++) {
            for (var x=0 ; x<this.gridWidth ; x++) {
                if (this._discovered[y][x] == 0) continue;  // Already discovered
                if (this._grid[y][x] == -1) continue;       // Mine

                for (var y2=Math.max(0, y-1) ; y2<=Math.min(this.gridHeight-1, y+1) ; y2++) {
                    for (var x2=Math.max(0, x-1) ; x2<=Math.min(this.gridWidth-1, x+1) ; x2++) {
                        if (this._discovered[y2][x2] == 0 && this._grid[y2][x2] == 0) {
                            this._discovered[y][x] = 0;
                            change = true;
                            break;
                        }
                    }
                }

            }
        }
        if (change) {
            this._expand();
        }
    },

    /**
     * Check if the player won.
     *
     * @method _checkWin
     * @private
     */
    _checkWin: function() {
        if (this.flags != this.nbMines) {
            return;
        }

        var match = 0;

        for (var y=0 ; y<this.gridHeight ; y++) {
            for (var x=0 ; x<this.gridWidth ; x++) {
                if (this._discovered[y][x] == 1 && this._grid[y][x] == -1) {
                    match++;
                }
            }
        }

        if (match == this.nbMines) {
            this.gameOver = true;
            this.win = true;
        }
    }

});


function main() {
    canvas = new photonui.Canvas();
    game = new Game(canvas);
    mouse = new photonui.MouseManager(canvas);
    mouse.threshold = 10;
    photonui.domInsert(canvas, document.getElementById("jsmined"));

    mouse.registerCallback("grid-click", "click", function(manager, mstate) {
        if (game.gameOver) {
            game = new Game(canvas);
            return;
        }

        var x = Math.floor(mstate.x/game.CELL_SIZE);
        var y = Math.floor(mstate.y/game.CELL_SIZE);
        if (mstate.button == "left") {
            game.discover(x, y);
        }
        else {
            game.toggleFlag(x, y);
        }
        game.refresh();

        if (game.gameOver) {
            if (game.win) {
                alert("Good Game"); //FIXME
            }
            else {
                alert("Boom!"); //FIXME
            }
        }
        else {
            console.log(game.flags + " / " + game.nbMines); //FIXME
        }
    });
}


var canvas = null;
var game = null;
var mouse = null;
var loader = new Loader();

// Load resources
loader.loadImage("bg-water", "assets/bg-water.png");
loader.loadImage("bg-flags", "assets/bg-flags.png");
loader.loadImage("bg-discovered", "assets/bg-discovered.png");
loader.loadImage("tileset", "assets/tileset.png");
loader.waitLoad(main);


